/*-------------------------------------------------------------------------+
| Copyright 2017 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.rcp.application.advisors;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.base.ui.utils.ConstraintsBaseUIUtils.activateConfiguration;
import static org.fortiss.tooling.base.ui.utils.ConstraintsBaseUIUtils.activateOneConfiguration;
import static org.fortiss.tooling.base.ui.utils.ConstraintsBaseUIUtils.deactivateConfiguration;
import static org.fortiss.tooling.base.ui.utils.ConstraintsBaseUIUtils.deactivateOneConfiguration;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ITreeSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CCombo;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.ISelectionService;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.menus.WorkbenchWindowControlContribution;
import org.fortiss.af3.project.model.IProjectConfigurationElement;
import org.fortiss.af3.project.model.configuration.DevelopmentProcessConfiguration;
import org.fortiss.af3.project.model.development.ConstraintBasedDevelopmentProcess;
import org.fortiss.af3.project.model.development.impl.DummyDevelopmentProcessImpl;
import org.fortiss.af3.project.model.impl.FileProjectImpl;
import org.fortiss.af3.project.model.impl.ProjectConfigurationImpl;
import org.fortiss.tooling.base.model.element.ConstraintConfiguration;
import org.fortiss.tooling.common.util.LambdaUtils;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceContainer;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.SuccessConstraintInstanceStatus;
import org.fortiss.tooling.kernel.ui.service.IConstraintUIService;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

// TODO (SB, 3)
// TODO (SB, 24): There seem to be many commonalities with CurrentObjectiveContributionItem -> derive base class and/or utility methods. Transfer my review and cleanup in CurrentObjectiveContributionItem also to this class! 
/**
 * @author levilucio
 */
public class UpcomingObjectiveContributionItem extends WorkbenchWindowControlContribution
		implements ISelectionListener, IResourceChangeListener {

	/** Helper to align the upcoming objective combo with the current objective combo. */
	public class SizedComposite extends Composite {
		/** Constructor. */
		public SizedComposite(Composite parent, int style) {
			super(parent, style);
		}

		/** {@inheritDoc} */
		@Override
		public void setSize(Point size) {
			super.setSize(computeSize(SWT.DEFAULT, SWT.DEFAULT, true));
		}

		/** {@inheritDoc} */
		@Override
		public void setSize(int width, int height) {
			Point size = computeSize(SWT.DEFAULT, SWT.DEFAULT, true);
			super.setSize(size.x, size.y);
		}

	}

	/** The initial upcoming objectives. */
	private String[] initialSetOfObjectives = {};

	/** Placeholder string for when no objective is selected. */
	private String initialObjective = "No upcoming objective available";

	/** The Combo holding the upcoming objectives. */
	private CCombo upcomingObjectiveCombo;

	/** Configuration of the current project selected in the combo box. */
	private ConstraintBasedDevelopmentProcess currentDevelopmentProcess = null;

	/** The current objective combo box. */
	CurrentObjectiveContributionItem currentObjective;

	/** Composite instance to hold several controls */
	private Composite composite;

	/** Constructor. */
	public UpcomingObjectiveContributionItem() {
		ISelectionService ww = org.eclipse.ui.PlatformUI.getWorkbench().getActiveWorkbenchWindow()
				.getSelectionService();
		ww.addSelectionListener(this);
		IWorkspace workspace = org.eclipse.core.resources.ResourcesPlugin.getWorkspace();
		workspace.addResourceChangeListener(this);
	}

	/** {@inheritDoc} */
	@Override
	public boolean isDynamic() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	protected Control createControl(Composite parent) {
		// This extra composite is only required here...
		composite = new SizedComposite(parent, SWT.NONE);
		GridLayout layout = new GridLayout(3, false);
		layout.marginWidth = 0;
		layout.verticalSpacing = 0;
		composite.setLayout(layout);

		// Add a separator to the current objective combo.
		new Label(composite, SWT.HORIZONTAL);

		Label label = new Label(composite, SWT.TRANSPARENT);
		label.setText("Upcoming Objective");

		// Objective Combo
		upcomingObjectiveCombo = new CCombo(composite, SWT.DROP_DOWN);
		upcomingObjectiveCombo.setItems(initialSetOfObjectives);
		upcomingObjectiveCombo.setText(initialObjective);
		upcomingObjectiveCombo.setToolTipText("Set the next modelling objective.");
		upcomingObjectiveCombo.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				currentObjective.updateObjectiveByName(((Combo)e.getSource()).getText());
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				// do nothing
			}
		});

		upcomingObjectiveCombo.addFocusListener(new FocusListener() {
			@Override
			public void focusGained(FocusEvent e) {
				if(currentDevelopmentProcess != null) {
					List<ConstraintConfiguration> upcomingConfigs = getUpcomingObjectives();

					if(upcomingConfigs.size() != 0) {
						List<String> configNames = new ArrayList<String>();

						for(ConstraintConfiguration constraintConfiguration : upcomingConfigs) {
							configNames.add(constraintConfiguration.getName());
						}

						String[] configNamesArray = new String[configNames.size()];
						configNames.toArray(configNamesArray);
						upcomingObjectiveCombo.setItems(configNamesArray);
						upcomingObjectiveCombo.setText(configNamesArray[0]);
					} else {
						upcomingObjectiveCombo.setItems(new String[0]);
						upcomingObjectiveCombo.setText(initialObjective);
					}
				}
			}

			@Override
			public void focusLost(FocusEvent e) {
				// do nothing
			}
		});

		parent.setSize(composite.computeSize(parent.getSize().x, SWT.DEFAULT));
		return composite;
	}

	/** Compute Combo Width */
	@Override
	protected int computeWidth(Control control) {
		return control.computeSize(SWT.DEFAULT, SWT.DEFAULT, true).x;
	}

	/** {@inheritDoc} */
	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		// set the current project to the one selected on the navigation tree

		if(selection instanceof ITreeSelection) {

			if(((ITreeSelection)selection).getFirstElement() != null) {
				Object objectChosenElement = ((ITreeSelection)selection).getFirstElement();

				if(objectChosenElement instanceof EObject) {
					EObject chosenElement = (EObject)objectChosenElement;

					FileProjectImpl fileProject = null;

					if(chosenElement instanceof FileProjectImpl) {
						fileProject = (FileProjectImpl)chosenElement;
					} else {
						fileProject = EcoreUtils.getFirstParentWithType(chosenElement,
								FileProjectImpl.class);
						if(fileProject == null) {
							// We are in the context of a Library.
							return;
						}
					}

					EList<IProjectRootElement> rootElems = fileProject.getRootElements();

					// there only exists one root for each project
					ProjectConfigurationImpl config = (ProjectConfigurationImpl)rootElems.get(0);
					EList<IProjectConfigurationElement> configElements =
							config.getConfigurationElements();

					DevelopmentProcessConfiguration processConfig =
							(DevelopmentProcessConfiguration)configElements.get(1);

					if(!(processConfig.getDevelopment() instanceof DummyDevelopmentProcessImpl)) {
						ConstraintBasedDevelopmentProcess process =
								(ConstraintBasedDevelopmentProcess)processConfig.getDevelopment();

						currentDevelopmentProcess = process;
					}
				}
				setUpcomingComboBasedOnCrieterion();
			}
		}
	}

	/**
	 * Find the upcoming objectives. These are the objectives following the ones in the
	 * dependence tree that have already been completely fulfilled (for which all the
	 * constraints are satisfied).
	 * 
	 * @return List of configurations not yet satisfied.
	 */
	private List<ConstraintConfiguration> getUpcomingObjectives() {
		// TODO: See #3246, 3248: This may or may not be the right solution
		if(currentDevelopmentProcess == null) {
			return emptyList();
		}
		EList<ConstraintConfiguration> configs = currentDevelopmentProcess.getConfigurations();
		List<ConstraintConfiguration> atomicConfigs =
				LambdaUtils.filterList(configs, c -> c.getIncludedConfigurations().isEmpty());

		List<ConstraintConfiguration> satisfiedObjectives = new ArrayList<>();
		List<ConstraintConfiguration> upcomingObjectives = new ArrayList<>();

		for(ConstraintConfiguration constraintConfiguration : atomicConfigs) {
			if(isConfigurationSatisfied(constraintConfiguration)) {
				satisfiedObjectives.add(constraintConfiguration);
			} else {
				upcomingObjectives.add(constraintConfiguration);
			}
		}

		boolean newSatisfiedAddedObjectives = !satisfiedObjectives.isEmpty();

		List<ConstraintConfiguration> newConfigsToTreat;

		while(newSatisfiedAddedObjectives) {

			newSatisfiedAddedObjectives = false;

			newConfigsToTreat = getDependentConfigurations(satisfiedObjectives);
			newConfigsToTreat.removeAll(upcomingObjectives);

			for(ConstraintConfiguration cstr : newConfigsToTreat) {
				if(isConfigurationSatisfied(cstr)) {
					satisfiedObjectives.add(cstr);
					newSatisfiedAddedObjectives = true;
				} else
					upcomingObjectives.add(cstr);
			}
		}

		return upcomingObjectives;
	}

	/**
	 * Find all configurations that depend on a set of given configurations.
	 * 
	 * @param configs
	 *            The configurations for which dependent configurations will be searched for.
	 * @return the configurations that depend on the input set of configuration.
	 */
	private List<ConstraintConfiguration>
			getDependentConfigurations(List<ConstraintConfiguration> configs) {
		List<ConstraintConfiguration> allConfigs = currentDevelopmentProcess.getConfigurations();
		ArrayList<ConstraintConfiguration> res = new ArrayList<ConstraintConfiguration>();
		for(ConstraintConfiguration cstr : allConfigs) {
			if(cstr.getIncludedConfigurations().size() != 0) {
				if(!configs.contains(cstr) &&
						configs.containsAll(cstr.getIncludedConfigurations())) {
					res.add(cstr);
				}
			}
		}
		return res;
	}

	/**
	 * Find if a given configuration is satisfied by the current state of the model.
	 * 
	 * @param constraintConfiguration
	 *            The configuration to check
	 * @return true if the configuration is satisfied, false otherwise
	 */
	private boolean isConfigurationSatisfied(ConstraintConfiguration constraintConfiguration) {
		ITopLevelElement modelContext = org.fortiss.tooling.kernel.service.IPersistencyService
				.getInstance().getTopLevelElementFor(currentDevelopmentProcess);

		boolean result = true;
		if(modelContext == null) {
			return result;
		}

		EList<String> activeConstraints = constraintConfiguration.getActiveConstraints();

		if(!activeConstraints.isEmpty()) {
			ConstraintConfiguration currentObjective =
					currentDevelopmentProcess.getCurrentObjective();
			IConstraintInstanceContainer constraintInstanceContainer =
					currentDevelopmentProcess.getConstraintInstanceContainer();
			modelContext.runAsNonDirtyingCommand(() -> {
				// deactivate current configurations
				deactivateConfiguration(currentObjective, constraintInstanceContainer);

				// activate configuration to be checked
				activateOneConfiguration(constraintConfiguration, constraintInstanceContainer);
			});

			for(String constraintName : activeConstraints) {
				EList<ConstraintInstance> instances =
						constraintInstanceContainer.getConstraintInstances();
				for(ConstraintInstance ci : instances) {
					if(constraintName.equalsIgnoreCase(ci.getConstraintName())) {
						IConstraintInstanceStatus status =
								IConstraintUIService.getInstance().getStatus(ci);
						if(!(status instanceof SuccessConstraintInstanceStatus)) {
							result = false;
							break;
						}
					}
				}
			}
			modelContext.runAsNonDirtyingCommand(() -> {

				// deactivate configuration to be checked
				deactivateOneConfiguration(constraintConfiguration, constraintInstanceContainer);

				// reactivate current configuration
				activateConfiguration(currentObjective, constraintInstanceContainer);
			});
		}

		return result;
	}

	/** Sets currentObjective. */
	public void setCurrentObjective(CurrentObjectiveContributionItem currentObjective) {
		this.currentObjective = currentObjective;
	}

	/** listener for resource changes */
	@Override
	public void resourceChanged(IResourceChangeEvent event) {
		if(event.getType() == org.eclipse.core.resources.IResourceChangeEvent.POST_CHANGE &&
				currentDevelopmentProcess != null && Display.getCurrent() != null) {
			setUpcomingComboBasedOnCrieterion();
		}
	}

	/**
	 * shows and hides the upcoming objective panel based on the number of upcoming objective
	 */
	private void setUpcomingComboBasedOnCrieterion() {
		if(getUpcomingObjectives().isEmpty() && composite != null) {
			upcomingObjectiveCombo.setEnabled(false);
		} else {
			upcomingObjectiveCombo.setEnabled(true);
			upcomingObjectiveCombo.setFocus();
		}

		upcomingObjectiveCombo.redraw();
	}
}
